unit Solar;

{$IMPORTEDDATA OFF}

interface

procedure SolarCalculate(Year, Month, Day: Integer;
    UTCOffset, Lat, Lon, Alt: Double;
    var SunriseHour, SunriseMin, SunsetHour, SunsetMin: Integer);

implementation

uses NMath;

function JuliusYear(Year, Month, Day, Hour, Min, Sec: Integer;
    UTCOffset: Double): Double;
begin
  Dec(Year, 2000);
  if Month <= 2 then
  begin
    Dec(Year);
    Inc(Month, 12);
  end;
  Result := (
      365*Year + 30*Month + Day - 33.5 - UTCOffset/24 + Floor(3*(Month+1)/5) +
      Floor(Year / 4) - Floor(Year / 100) + Floor(Year / 400) +
      Hour / 24 + Min / 1440 + Sec / 86400 +
      (65 + Year) / 86400
      ) / 365.25;
end;

function SiderealTime(JY: Double; Sec: Integer; UTCOffset, Lon: Double): Double;
begin
  Result := FMod(
      100.4606 + 360.007700536 * JY + 0.00000003879 * JY * JY - 15*UTCOffset +
      Lon + 360 * Sec / 86400,
      360);
end;

function SolarCelestialLongitude(JY: Double): Double;
begin
  Result := FMod(
      280.4603 + 360.00769 * JY +
      (1.9146 - 0.00005 * JY) * Sin((357.538 + 359.991 * JY) * Pi/180) +
      0.0200 * Sin((355.05 + 719.981 * JY) * Pi/180) +
      0.0048 * Sin((234.95 + 19.341 * JY) * Pi/180) +
      0.0020 * Sin((247.1 + 329.640 * JY) * Pi/180) +
      0.0018 * Sin((297.8 + 4452.67  * JY) * Pi/180) +
      0.0018 * Sin((251.3 + 0.20  * JY) * Pi/180) +
      0.0015 * Sin((343.2 + 450.37 * JY) * Pi/180) +
      0.0013 * Sin((81.4 + 225.18 * JY) * Pi/180) +
      0.0008 * Sin((132.5 + 659.29 * JY) * Pi/180) +
      0.0007 * Sin((153.3 + 90.38 * JY) * Pi/180) +
      0.0007 * Sin((206.8 + 30.35 * JY) * Pi/180) +
      0.0006 * Sin((29.8 + 337.18 * JY) * Pi/180) +
      0.0005 * Sin((207.4 + 1.50 * JY) * Pi/180) +
      0.0005 * Sin((291.2 + 22.81 * JY) * Pi/180) +
      0.0004 * Sin((234.9 + 315.56 * JY) * Pi/180) +
      0.0004 * Sin((157.3 + 299.30 * JY) * Pi/180) +
      0.0004 * Sin((21.1 + 720.02 * JY) * Pi/180) +
      0.0003 * Sin((352.5 + 1079.97 * JY) * Pi/180) +
      0.0003 * Sin((329.7 + 44.43 * JY) * Pi/180),
      360
      );
end;

function SolarDistance(JY: Double): Double;
begin
  // unit: AU
  Result := Pow(10,
      (0.007256 - 0.0000002 * JY) * Sin((267.54 + 359.991 * JY) * Pi/180) +
      0.000091 * Sin((265.1 +  719.98 * JY) * Pi/180) +
      0.000030 * Sin(90.0 * Pi/180) +
      0.000013 * Sin((27.8 + 4452.67 * JY) * Pi/180) +
      0.000007 * Sin((254 + 450.4 * JY) * Pi/180) +
      0.000007 * Sin((156 + 329.6 * JY) * Pi/180)
      );
end;

function SolarDeclination(JY, Lon: Double): Double;
begin
  Result :=
      ArcTan(Tan(Lon*Pi/180) * Cos((23.439291 - 0.000130042 * JY)*Pi/180))
      * 180/Pi;
  if (Lon >= 0) and (Lon < 180) then
  begin
   while Result < 0 do Result := Result + 180;
   while Result >= 180 do Result := Result - 180;
  end
  else begin
   while Result < 180 do Result := Result + 180;
   while Result >= 360 do Result := Result - 180;
  end;
end;

function SolarRightAscension(JY, Lon: Double): Double;
begin
  Result :=
      ArcSin(Sin(Lon*Pi/180) * Sin((23.439291 - 0.000130042 * JY)*Pi/180))
      * 180/Pi;
end;

function SolarAltitude(Lat, Sidereal, Decl, RA: Double): Double;
begin
  Result := ArcSin(
    Sin(RA*Pi/180) * Sin(Lat*Pi/180) +
    Cos(RA*Pi/180) * Cos(Lat*Pi/180) * Cos((Sidereal - Decl) * Pi/180)
    ) * 180/Pi;
end;

function ApparentHorizonAltitude(Alt, Dist: Double): Double;
begin
  Result := (0.002442818 / Dist - 0.035333333 * Sqrt(Alt)) -
      0.26699444 / Dist - 0.585555555;
end;

procedure SolarCalculate(Year, Month, Day: Integer;
    UTCOffset, Lat, Lon, Alt: Double;
    var SunriseHour, SunriseMin, SunsetHour, SunsetMin: Integer);
const
  step = 30; // sec.
var
  jy, slon, sa, psa, t: Double;
  s: Integer;
begin
  SunriseHour := -1;
  SunsetHour := -1;

  jy := JuliusYear(Year, Month, Day-1, 23, 59, 60-step, UTCOffset);
  slon := SolarCelestialLongitude(jy);
  psa := SolarAltitude(
    Lat,
    SiderealTime(jy, 0, UTCOffset, Lon),
    SolarDeclination(jy, slon),
    SolarRightAscension(jy, slon)
    );

  jy := JuliusYear(Year, Month, Day, 0, 0, 0, UTCOffset);
  s := 0;
  while s < 24*60*60 do
  begin
    slon := SolarCelestialLongitude(jy);
    sa := SolarAltitude(
      Lat,
      SiderealTime(jy, s, UTCOffset, Lon),
      SolarDeclination(jy, slon),
      SolarRightAscension(jy, slon)
      );
    t := ApparentHorizonAltitude(Alt, SolarDistance(jy));
    if (psa < t) and (sa >= t) then
    begin
      SunriseHour := s div 3600 ;
      SunriseMin := s mod 3600 div 60;
    end
    else if (psa > t) and (sa <= t) then
    begin
      SunsetHour := s div 3600;
      SunsetMin := s mod 3600 div 60;
    end;
    if (SunriseHour > -1) and (SunsetHour > -1) then
      Exit;
    psa := sa;
    jy := jy + (step/60/1440) / 365.25;
    Inc(s, step);
  end;
end;

end.
